/*****************************************************************************/
/*!
 * \file theory_core.h
 * 
 * Author: Clark Barrett
 * 
 * Created: Thu Jan 30 16:58:05 2003
 *
 * <hr>
 * Copyright (C) 2003 by the Board of Trustees of Leland Stanford
 * Junior University and by New York University. 
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.  In particular:
 *
 * - The above copyright notice and this permission notice must appear
 * in all copies of the software and related documentation.
 *
 * - THE SOFTWARE IS PROVIDED "AS-IS", WITHOUT ANY WARRANTIES,
 * EXPRESSED OR IMPLIED.  USE IT AT YOUR OWN RISK.
 * 
 * <hr>
 * 
 */
/*****************************************************************************/

#ifndef _cvcl__include__theory_core_h_
#define _cvcl__include__theory_core_h_

#include "debug.h"
#include "theory.h"
#include <queue>
#include "compat_hash_map.h"
#include "expr_map.h"
#include "cdo.h"
#include "cdlist.h"
#include "cdmap.h"

namespace CVCL {

class TheoremManager;
class CommonProofRules;
class SearchEngine;

/*****************************************************************************/
/*!
 *\class TheoryCore
 *\ingroup Theories
 *\brief This theory handles the built-in logical connectives plus equality.
 * It also handles the registration and cooperation among all other theories.
 *
 * Author: Clark Barrett
 *
 * Created: Sat Feb 8 14:54:05 2003
 */
/*****************************************************************************/
class TheoryCore :public Theory {
  friend class Theory;

  //! Expression manager
  ExprManager* d_em;

  //! PrettyPrinter (we own it)
  PrettyPrinter* d_printer;

  //! Search Engine
  VCL* d_vcl;

  //! Assertion queue
  std::queue<Theorem> d_queue;
  //! Queue of facts to be sent to the SearchEngine
  std::vector<Theorem> d_queueSE;

  //! Equality queue
  /*!
   * Contains Theorems in the form of e1==e2, where e2 is i-leaf
   * simplified in the current context.  
   *
   * \sa enqueueEquality().
   */
  std::vector<Theorem> d_equalityQueue;

  //! Inconsistent flag
  CDO<bool> d_inconsistent;
  //! The set of reasons for incompleteness (empty when we are complete)
  CDMap<std::string, bool> d_incomplete;
  //! List of known disequalities (to be processed at every checkSat() call)
  CDList<Theorem> d_diseq;

  //! Proof of inconsistent
  CDO<Theorem> d_incThm;
  //! List of all active terms in the system (for quantifier instantiation)
  CDList<Expr> d_terms;
  //! List of variables that were created up to this point
  CDList<Expr> d_vars;
  //! Database of declared identifiers
  std::map<std::string, Expr> d_globals;
  //! Bound variable stack: a vector of pairs <name, var>
  std::vector<std::pair<std::string, Expr> > d_boundVarStack;

  //! List of all terms that are shared between theories (alien subterms)
  /*! Maps each shared term to its own theory. */
  CDMap<Expr, Theory*> d_sharedTerms;

  //! Array of registered theories
  std::vector<Theory*> d_theories;

  //! Array mapping kinds to theories
  std::map<int, Theory*> d_theoryMap;

  //! The theory which has the solver (if any)
  Theory* d_solver;

  //! Mapping of compound type variables to simpler types (model generation)
  ExprHashMap<std::vector<Expr> > d_varModelMap;
  //! Mapping of intermediate variables to their valies
  ExprHashMap<Theorem> d_varAssignments;
  //! List of basic variables (temporary storage for model generation)
  std::vector<Expr> d_basicModelVars;
  //! Mapping of basic variables to simplified expressions (temp. storage)
  /*! There may be some vars which simplify to something else; we save
   * those separately, and keep only those which simplify to
   * themselves.  Once the latter vars are assigned, we'll re-simplify
   * the former variables and use the results as concrete values.
  */
  ExprHashMap<Theorem> d_simplifiedModelVars;

  //! Command line flag whether to simplify in place
  const bool* d_simplifyInPlace;

  //! Command line flag whether to convert to CNF
  const bool* d_cnfOption;

  //! Cache for pushNegation()
  ExprMap<Theorem> d_pushNegCache;

  //! Resource limit (0==unlimited, 1==no more resources, >=2 - available).
  /*! Currently, it is the number of enqueued facts.  Once the
   * resource is exhausted, the remaining facts will be dropped, and
   * an incomplete flag is set. 
   */
  unsigned d_resourceLimit;

  //! Command line flag whether to dump debugging info on every new fact
  IF_DEBUG(const std::string* d_dumpTrace);
  IF_DEBUG(bool d_inCheckSATCore);
  IF_DEBUG(bool d_inAddFact);
  IF_DEBUG(bool d_inSimplify);

  //! So we get notified every time there's a pop
  friend class CoreNotifyObj;
  class CoreNotifyObj :public ContextNotifyObj {
    TheoryCore* d_theoryCore;
  public:
    CoreNotifyObj(TheoryCore* tc, Context* context)
      : ContextNotifyObj(context), d_theoryCore(tc) {}
    void notify() { d_theoryCore->d_em->invalidateSimpCache(); }
  };
  CoreNotifyObj d_notifyObj;

private:
  //! Set the find pointer of an atomic formula and notify SearchEngine
  /*! \param thm is a Theorem(phi) or Theorem(NOT phi), where phi is
   * an atomic formula to get a find pointer to TRUE or FALSE,
   * respectively.
   * \param notifySAT indicates whether to notify the Search Engine of
   * this literal.
   */
  void setFindLiteral(const Theorem& thm, bool notifySAT);
  //! Auxiliary recursive function for pushNegation().
  Theorem pushNegationRec(const Expr& e, bool neg);
  //! Its version for transitivity
  Theorem pushNegationRec(const Theorem& e, bool neg);
  //! Derived rule for rewriting ITE
  Theorem rewriteIte(const Expr& e);
  //! Core rewrites for literals (NOT and EQ)
  Theorem rewriteLitCore(const Expr& e);
  /*! @brief An auxiliary function for assertEqualities(); return true
   *  if inconsistency is derived.
   */
  bool processEquality(const Theorem& thm, ExprMap<Theorem>& q);
  //! Auxiliary function for transformDiseq()
  Theorem traverseDiseq(const Expr& lhs, const Expr& rhs,
			ExprMap<Theorem>& cache, std::vector<Expr>& eqs);
  //! Transform disequalities using structural similarities
  /*!
   * e.g.: f(x,g(y)) /= f(z,g(w)) is turned into (x /= z) OR (y /= w)
   */
  Theorem transformDiseq(const Theorem& thm);
  //! Private method to get a new theorem producer.
  /*! This method is used ONLY by the TheoryCore constructor.  The
    only reason to have this method is to separate the trusted part of
    the constructor into a separate .cpp file (Alternative is to make
    the entire constructer trusted, which is not very safe). */
  CommonProofRules* newCommonRules(TheoremManager* tm);
  //! Enqueue a fact to be sent to the SearchEngine
  void enqueueSE(const Theorem& thm);
  //! Fetch the concrete assignment to the variable during model generation
  Theorem getModelValue(const Expr& e);

public:
  //! Constructor
  TheoryCore(VCL *vcl);
  //! Destructor
  ~TheoryCore();

  /*************************************************************************/
  /*!
   *\name Preprocessor methods
   * FIXME: please document these functions
   *@{
   */
  /*************************************************************************/
  Theorem preprocess(const Expr& e);
  Theorem preprocess(const Theorem& thm);
  // Push all negations down to the leaves
  Theorem pushNegation(const Expr& e);
  Theorem substitute(Expr e, ExprMap<Theorem> *init_st=NULL);
  Theorem ite_simplify(Expr e);
  Theorem ite_convert(const Expr& e);
  Expr ite_reorder(const Expr& e);
  Expr getNeg(const Expr& e);
  /*@}*/ // end of preprocessor stuff

  // Theory interface for basic kinds
  /*! Variables of uninterpreted types may be sent here, and they
    should be ignored. */
  void addSharedTerm(const Expr& e) { }
  void assertFact(const Theorem& e);
  void checkSat(bool fullEffort);
  void computeModelBasic(const std::vector<Expr>& v);
  Theorem rewrite(const Expr& e);
  Theorem simplifyOp(const Expr& e);
  void setup(const Expr& e);
  void update(const Theorem& e, const Expr& d) {DebugAssert(false,""); }
  Expr computeTypePred(const Type& t,const Expr& e);
  void computeType(const Expr& e);
  Type computeBaseType(const Type& t);
  Expr computeTCC(const Expr& e);
  ExprStream& print(ExprStream& os, const Expr& e);
  Expr parseExpr(const Expr& e);
  //! Top-level call to parseExpr, clears the bound variable stack.
  /*! Use it in VCL instead of parseExpr(). */
  Expr parseExprTop(const Expr& e) {
    d_boundVarStack.clear();
    return parseExpr(e);
  }
  //! An auxiliary recursive function to process COND expressions into ITE
  Expr processCond(const Expr& e, int i);
  Expr parseExprOp(const Expr& e);


  //! Adds expression to var database
  void addToVarDB(const Expr & e);
  //! Split compound vars into basic type variables
  /*! The results are saved in d_basicModelVars and
   *  d_simplifiedModelVars.  Also, add new basic vars as shared terms
   *  whenever their type belongs to a different theory than the term
   *  itself.
   */
  void collectBasicVars();
  //! Calls theory specific computeModel, results are placed in map
  void buildModel(ExprMap<Expr>& m);
  //! Recursively build a compound-type variable assignment for e
  void collectModelValues(const Expr& e,
			  const ExprHashMap<Theorem>& simplifiedVars,
			  ExprMap<Expr>& m);
  //! Asserts the disequalities of all shared terms. 
  void assertDisequalities();
  //! Calls for other theories to add facts to refine a coutnerexample.
  void refineCounterExample();

  // The remaining functions implement pseudocode described in my Ph.D. thesis
  // -- Clark.  They have been extended to use theorems (rather than just
  // Expr's) to track the chain of reasoning used in the transformations.

  //! Add an new fact to the core
  void addFact(const Theorem& e);
  //! Add an equality to the core (for use in model generation)
  /*! Use with caution.  The theorem 'e' must be an equality with the
   *  RHS fully i-leaf-simplified, and other invariants of
   *  assertEqualities() must be satisfied.
   */
  void addEquality(const Theorem& e);
  //! A helper function for addFact()
  void processFactQueue();
  //! Top-level simplifier
  Theorem simplifyThm(const Expr& e, bool forceRebuild = false);
  bool inconsistent() { return d_inconsistent ; }
  bool incomplete() { return d_incomplete.size() > 0 ; }
  bool incomplete(std::vector<std::string>& reasons);
  Theorem inconsistentThm() { return d_incThm; } 
  /*! @brief Returns phi |= e = rewriteCore(e).  "Core" added to avoid
    conflict with theory interface function name */
  Theorem rewriteCore(const Expr& e);
  Theorem rewriteLiteral(const Expr& e);
  /*! @brief To be called by SearchEngine when it believes the context
   * is satisfiable.
   *
   * \return true if no more facts are found, and false if more
   * processing needs to be done.
   */
  virtual bool checkSATCore();

  //! Test whether a kind maps to any theory
  bool hasTheory(int kind);

  // Must provide proof of e from H and phi
  //! Enqueue a new fact
  void enqueueFact(const Theorem& e);
  //! Enqueue a new equality
  void enqueueEquality(const Theorem& e);
  //! Suggest a splitter to the SearchEngine
  void addSplitter(const Expr& e, int priority = 0);
  // Must provide proof of inconsistency
  void setInconsistent(const Theorem& e);
  void setIncomplete(const std::string& reason);
  void setupTerm(const Expr& e, Theory* i);
  //! Set the resource limit (0==unlimited).
  void setResourceLimit(unsigned limit);

  const CDList<Expr>& getTerms() { return d_terms; }

  //! Compute and cache the subtyping predicates for the expression
  /*!
   * Note: e caches the conjunction of <em>all</em> predicates for its
   * subexpressions.  So, when doing a look-up, it is sufficient to
   * assert just the predicate for the top-level e, without traversing
   * e recursively.
   */
  Theorem subtypePredicate(const Expr& e);

  //! Check whether e is a propositional atom (not a Boolean connective)
  bool isPropAtom(const Expr& e);
  //! Check whether e is a propositional literal (prop atom or its negation)
  bool isPropLiteral(const Expr& e);

 private:
  /*! @brief The assumptions for e must be in H or phi.  "Core" added
   * to avoid conflict with theory interface function name
   */
  void assertFactCore(const Theorem& e);
  //! Assert a system of equations (1 or more).
  /*! e is either a single equation, or a conjunction of equations
   */
  void assertEqualities(const Theorem& e);
  //! Process a newly derived formula
  void assertFormula(const Theorem& e);
  //! The recursive simplifier (to be used only in DP-specific simplifyOp())
  Theorem simplify(const Expr& e);
  //! The recursive simplifier with the in-place rewriting optimization
  Theorem simplifyInPlaceThmRec(const Expr& e);
  //! Transitive composition of other rewrites with the above
  Theorem rewriteCore(const Theorem& e);
  /*! @brief Given a proof of e, returns a proof of the existential
   * closure of the solved form of e
   */
  Theorem solve(const Theorem& e);
  //! Implementation of the existential elimination (skolemization) rule
  Theorem skolemize(const Theorem& thm);
  //! Returns a theorem |- e = v for a new Skolem variable v
  Theorem varIntroSkolem(const Expr& e);
  //! Assign t a concrete value val.  Used in model generation.
  void assignValue(const Expr& t, const Expr& val);
  //! Record a derived assignment to a variable (LHS).
  void assignValue(const Theorem& thm);
  //! Request a unit of resource
  /*! It will be subtracted from the resource limit. 
   *
   * \return true if resource unit is granted, false if no more
   * resources available.
   */
  bool getResource();
};

}

#endif
