/*****************************************************************************/
/*!
 * \file arith_expr.h
 * 
 * Author: Sergey Berezin
 * 
 * Created: Mon Apr 18 11:35:12 2005
 *
 * <hr>
 * Copyright (C) 2003 by the Board of Trustees of Leland Stanford
 * Junior University and by New York University. 
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.  In particular:
 *
 * - The above copyright notice and this permission notice must appear
 * in all copies of the software and related documentation.
 *
 * - THE SOFTWARE IS PROVIDED "AS-IS", WITHOUT ANY WARRANTIES,
 * EXPRESSED OR IMPLIED.  USE IT AT YOUR OWN RISK.
 * 
 * <hr>
 * 
 */
/*****************************************************************************/

#ifndef _CVCL__theory_arith__arith_expr_h_
#define _CVCL__theory_arith__arith_expr_h_

#include "expr_value.h"

namespace CVCL {

  //! Subclass of ExprValue for gray shadow expressions
  /*! GRAY_SHADOW(v, e, c1, c2) represents OR(v = e + i)_{c1 <= i <= c2}
   */
  class ExprGrayShadow: public ExprValue {
  private:
    size_t d_mmIdx;
    Expr d_v;
    Expr d_e;
    Rational d_c1, d_c2;
    
    // Hash function for Rational
    static size_t hash(const Rational& r) {
      return s_charHash(r.toString().c_str());
    }
  protected:
    size_t computeHash() const {
      size_t res(d_v.hash());
      res = PRIME*res + d_e.hash();
      res = PRIME*res + hash(d_c1);
      res = PRIME*res + hash(d_c2);
      return res;
    }
  public:
    // Constructor
    ExprGrayShadow(ExprManager* em, const Expr& v, const Expr& e,
		   const Rational& c1, const Rational& c2,
		   size_t mmIndex, ExprIndex idx=0)
      : ExprValue(em, GRAY_SHADOW, idx), d_mmIdx(mmIndex),
      d_v(v), d_e(e), d_c1(c1), d_c2(c2) { }
    // Destructor
    ~ExprGrayShadow() { }
    
    size_t getMMIndex() const { return d_mmIdx; }
    
    bool operator==(const ExprValue& ev2) const {
      if(d_mmIdx != ev2.getMMIndex()) return false;
      return (getKind() == ev2.getKind()
	      && getExprValue(0) == ev2.getExprValue(0)
	      && getExprValue(1) == ev2.getExprValue(1)
	      && getRatValue(0) == ev2.getRatValue(0)
	      && getRatValue(1) == ev2.getRatValue(1));
    }
    
    ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const {
      return new(em->getMM(getMMIndex()))
	ExprGrayShadow(em, d_v, d_e, d_c1, d_c2, d_mmIdx, idx);
    }

    // Accessing the values
    const Rational& getRatValue(int idx) const {
      DebugAssert(0 <= idx && idx < 2, "ExprGrayShadow::getRatValue("
		  +int2string(idx)+"): out of bounds");
      if(idx==0) return d_c1;
      else return d_c2;
    }

    const Expr& getExprValue(int idx) const {
      DebugAssert(0<= idx && idx <= 1, "ExprGrayShadow::getExprValue("
		  +int2string(idx)+"): out of bounds");
      if(idx==0) return d_v;
      else return d_e;
    }

    size_t getRatValueSize() const { return 2; }
    size_t getExprValueSize() const { return 2; }
    
    
    // Memory management
    void* operator new(size_t size, MemoryManager* mm) {
      return mm->newData(size);
    }
    void operator delete(void*) { }
  }; // end of class ExprGrayShadow
  
} // end of namespace CVCL

#endif
