/*****************************************************************************/
/*!
 * \file common_theorem_producer.cpp
 * 
 * Author: Sergey Berezin
 * 
 * Created: Feb 05 03:40:36 GMT 2003
 *
 * <hr>
 * Copyright (C) 2003 by the Board of Trustees of Leland Stanford
 * Junior University and by New York University. 
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.  In particular:
 *
 * - The above copyright notice and this permission notice must appear
 * in all copies of the software and related documentation.
 *
 * - THE SOFTWARE IS PROVIDED "AS-IS", WITHOUT ANY WARRANTIES,
 * EXPRESSED OR IMPLIED.  USE IT AT YOUR OWN RISK.
 * 
 * <hr>
 * 
 */
/*****************************************************************************/
// CLASS: CommonTheoremProducer
//
// AUTHOR: Sergey Berezin, 12/09/2002
//         Vijay Ganesh, september 15th, 2004 (CNF Converter rules)
//
// Description: Implementation of the proof rules for ground
// equational logic (reflexivity, symmetry, transitivity, and
// substitutivity).
//
///////////////////////////////////////////////////////////////////////////////

// This code is trusted
#define _CVCL_TRUSTED_

#include "common_theorem_producer.h"
#include "theory_core.h"
#include "vcl.h"
#include "debug.h"

using namespace CVCL;
using namespace std;

// The trusted method of TheoryCore which creates this theorem producer
CommonProofRules* 
TheoryCore::newCommonRules(TheoremManager* tm) {
  return new CommonTheoremProducer(tm, this);
}

// ==> u:a |- a

// The label 'u' is generated to be unique, and it becomes the proof
// of the assumption.  Thus, there is no special rule for generating
// assumptions in the proof system, there is a family of dynamically
// generated "axioms" instead.
Theorem
CommonTheoremProducer::assumpRule(const Expr& e) {
  TRACE("axioms", "assumpRule(", e, ")");
  Proof pf;
  if(withProof()) pf = newLabel(e);
  return newAxiom(e, pf);
}


//  ==> a == a  or  ==> a IFF a
Theorem
CommonTheoremProducer::reflexivityRule(const Expr& a) {
  Proof pf;
  if(withProof()) {
    Type t = getType(a);
    if(CHECK_PROOFS)
      CHECK_SOUND(!t.isNull(),
		  "reflexivityRule: 'a' has no type: " + a.toString());
    bool useIff = (t == d_tm->getVCL()->boolType());
    pf = newPf(useIff? "iff_refl" : "eq_refl", t.getExpr(), a);
  }
  return newReflTheorem(a, pf);
}

//  a1 == a2 ==> a2 == a1 (same for IFF)
Theorem
CommonTheoremProducer::symmetryRule(const Theorem& a1_eq_a2) {
  if(CHECK_PROOFS)
    CHECK_SOUND(a1_eq_a2.isRewrite(),
		("CVCL::CommonTheoremProducer: "
		 "theorem is not an equality or iff:\n  "
		 + a1_eq_a2.getExpr().toString()).c_str());
  const Expr& a1 = a1_eq_a2.getLHS();
  const Expr& a2 = a1_eq_a2.getRHS();

  Proof pf;
  Assumptions a;
  /////////////////////////////////////////////////////////////////////////
  //// Proof compaction
  /////////////////////////////////////////////////////////////////////////
  // If a1 == a2, use reflexivity
  if(a1 == a2) return reflexivityRule(a1);
  // Otherwise, do the work
  if (withAssumptions()) a = a1_eq_a2.getAssumptionsCopy();
  if(withProof()) {
    Type t = getType(a1);
    // Check the types
    IF_DEBUG(getType(a1_eq_a2.getExpr()));
    bool isEquality = !t.isBool();
    if(isEquality) {
      vector<Expr> args;
      args.push_back(t.getExpr());
      args.push_back(a1);
      args.push_back(a2);
      pf = newPf("eq_symm", args, a1_eq_a2.getProof());
    } else
      pf = newPf("iff_symm", a1, a2, a1_eq_a2.getProof());
  }
  return newRWTheorem(a2, a1, a, pf);
}

// ==> (a1 == a2) IFF (a2 == a1) [same for IFF]
Theorem
CommonTheoremProducer::rewriteUsingSymmetry(const Expr& a1_eq_a2) {
  bool isIff = a1_eq_a2.isIff();
  if(CHECK_PROOFS)
    CHECK_SOUND(a1_eq_a2.isEq() || isIff, "rewriteUsingSymmetry precondition violated");
  const Expr& a1 = a1_eq_a2[0];
  const Expr& a2 = a1_eq_a2[1];
  // Proof compaction: if a1 == a2, use reflexivity
  if(a1 == a2) return reflexivityRule(a1_eq_a2);
  // Otherwise, do the work
  Assumptions assump; // No assumptions, leave it Null
  Proof pf;
  if(withProof()) {
    Type t = getType(a1);
    DebugAssert(!t.isNull(),
		"rewriteUsingSymmetry: a1 has no type: " + a1.toString());
    if(isIff)
      pf = newPf("rewrite_iff_symm", a1, a2);
    else {
      pf = newPf("rewrite_eq_symm", t.getExpr(), a1, a2);
    }
  }
  return newRWTheorem(a1_eq_a2, isIff ? a2.iffExpr(a1) : a2.eqExpr(a1), assump, pf);
}

// (a1 == a2) & (a2 == a3) ==> (a1 == a3)   [same for IFF]
Theorem
CommonTheoremProducer::transitivityRule(const Theorem& a1_eq_a2,
					const Theorem& a2_eq_a3) {
  DebugAssert(!a1_eq_a2.isNull(), "transitivityRule()");
  DebugAssert(!a2_eq_a3.isNull(), "transitivityRule()");
  if(CHECK_PROOFS) {
    CHECK_SOUND(a1_eq_a2.isRewrite() && a2_eq_a3.isRewrite(),  
		"CVCL::CommonTheoremProducer::transitivityRule:\n  "
		"Wrong premises: first = "
		+ a1_eq_a2.getExpr().toString() + ", second = "
		+ a2_eq_a3.getExpr().toString());
    CHECK_SOUND(a1_eq_a2.getRHS() == a2_eq_a3.getLHS(),
		"CVCL::CommonTheoremProducer::transitivityRule:\n  "
		"Wrong premises: first = "
		+ a1_eq_a2.getExpr().toString() + ", second = "
		+ a2_eq_a3.getExpr().toString());
  }
  const Expr& a1 = a1_eq_a2.getLHS();
  const Expr& a2 = a1_eq_a2.getRHS();
  const Expr& a3 = a2_eq_a3.getRHS();

  /////////////////////////////////////////////////////////////////////////
  //// Proof compaction
  /////////////////////////////////////////////////////////////////////////
  // if a1 == a3, use reflexivity (and lose all assumptions)
  if(a1 == a3) return reflexivityRule(a1);
  // if a1 == a2, or if a2 == a3, use only the non-trivial part
  if(a1 == a2) return a2_eq_a3;
  if(a2 == a3) return a1_eq_a2;

  ////////////////////////////////////////////////////////////////////////
  //// No shortcuts.  Do the work.
  ////////////////////////////////////////////////////////////////////////

  Assumptions a;
  Proof pf;
  // Merge the assumptions
  if(withAssumptions())
    a = merge(a1_eq_a2, a2_eq_a3);
  // Build the proof
  if(withProof()) {
    Type t = getType(a1);
    bool isEquality = (t != d_tm->getVCL()->boolType());
    string name((isEquality)? "eq_trans" : "iff_trans");
    vector<Expr> args;
    vector<Proof> pfs;
    DebugAssert(!t.isNull(), "transitivityRule: "
		"Type is not computed for a1: " + a1.toString());
    // Type argument is needed only for equality
    if(isEquality) args.push_back(t.getExpr());
    args.push_back(a1);
    args.push_back(a2);
    args.push_back(a3);
    pfs.push_back(a1_eq_a2.getProof());
    pfs.push_back(a2_eq_a3.getProof());
    pf = newPf(name, args, pfs);
  }
  return newRWTheorem(a1, a3, a, pf);
}
// (c_1 == d_1) & ... & (c_n == d_n)
//   ==> op(c_1,...,c_n) == op(d_1,...,d_n)
Theorem
CommonTheoremProducer::substitutivityRule(const Op& op,
					  const vector<Theorem>& thms) {
  IF_DEBUG
    (static DebugTimer
       accum0(debugger.timer("substitutivityRule time"));
     static DebugTimer tmpTimer(debugger.newTimer());
     static DebugCounter count(debugger.counter("substitutivityRule calls"));
     debugger.setCurrentTime(tmpTimer);
     count++);
  // Check for trivial case: no theorems, return (op == op)
  unsigned size(thms.size());
  if(size == 0)
    return reflexivityRule(newExpr(d_tm->getEM(), op));
  // Check that theorems are of the form c_i == d_i and collect
  // vectors of c_i's and d_i's and the vector of proofs
  vector<Expr> c, d;
  vector<Proof> pfs;
  // Reserve memory for argument vectors.  Do not reserve memory for
  // pfs, they are rarely used and slow anyway.
  c.reserve(size); d.reserve(size);
  for(vector<Theorem>::const_iterator i = thms.begin(), iend = thms.end();
      i != iend; ++i) {
    // Check that t is c == d or c IFF d
    if(CHECK_PROOFS)
      CHECK_SOUND(i->isRewrite(),
		  "CVCL::CommonTheoremProducer::substitutivityRule:\n  "
		  "premis is not an equality or IFF: "
		  + i->getExpr().toString()
		  + "\n  op = " + op.toString());
    // Collect the pieces
    c.push_back(i->getLHS());
    d.push_back(i->getRHS());
    if(withProof()) pfs.push_back(i->getProof());
  }
  Expr e1(op, c), e2(op, d);
  // Proof compaction: if e1 == e2, use reflexivity
  if(e1 == e2) return reflexivityRule(e1);
  // Otherwise, do the work
  Assumptions a;
  Proof pf;
  if(withAssumptions()) a = merge(thms);
  if(withProof())
    // FIXME: this rule is not directly expressible in flea
    pf = newPf("basic_subst_op",e1,e2,pfs);
  Theorem res = newRWTheorem(e1, e2, a, pf);
  IF_DEBUG(debugger.setElapsed(tmpTimer);
	   accum0 += tmpTimer);
  return res;
}

// (c_1 == d_1) & ... & (c_n == d_n)
//   ==> op(c_1,...,c_n) == op(d_1,...,d_n)
Theorem
CommonTheoremProducer::substitutivityRule(const Expr& e,
					  const vector<unsigned>& changed,
					  const vector<Theorem>& thms) {
//   if(true){
//     Op op = Op(e.getKind());
//     return substitutivityRule(op,thms); //XX    
//   }
  IF_DEBUG
    (static DebugTimer
       accum0(debugger.timer("substitutivityRule2 time"));
     static DebugTimer tmpTimer(debugger.newTimer());
     static DebugCounter count(debugger.counter("substitutivityRule2 calls"));
     debugger.setCurrentTime(tmpTimer);
     count++);
  unsigned csize = changed.size();
  DebugAssert(csize > 0, "substitutivityRule2 should not be called");
  DebugAssert(csize == thms.size(), "substitutivityRule2: wrong args");

  // Check that theorems are of the form c_i == d_i and collect
  // vectors of c_i's and d_i's and the vector of proofs
  unsigned size = e.arity();
  vector<Expr> c;
  bool checkProofs(CHECK_PROOFS);
  for(unsigned j = 0, k = 0; k < size; ++k) {
    if (j == csize || changed[j] != k) {
      c.push_back(e[k]);
      continue;
    }
    // Check that t is c == d or c IFF d
    if(checkProofs)
      CHECK_SOUND(thms[j].isRewrite() && thms[j].getLHS() == e[k],
		  "CVCL::CommonTheoremProducer::substitutivityRule:\n  "
		  "premis is not an equality or IFF: "
		  + thms[j].getExpr().toString()
		  + "\n  e = " + e.toString());
    // Collect the pieces
    c.push_back(thms[j].getRHS());
    ++j;
  }
  Expr e2(getOp(e), c);
  IF_DEBUG(if(e == e2) {
    ostream& os = debugger.getOS();
    os << "substitutivityRule2: e = " << e << "\n e2 = " << e2
       << "\n changed kids: [\n";
    for(unsigned j=0; j<changed.size(); j++)
      os << "  (" << changed[j] << ") " << thms[j] << "\n";
    os << "]\n";
  });
  DebugAssert(e != e2,
	      "substitutivityRule2 should not be called in this case:\n"
	      "e = "+e.toString());

  Assumptions a;
  Proof pf;
  if(withAssumptions()) a = merge(thms);
  vector<Proof> pfs;
  if(withProof())
    for(vector<Theorem>::const_iterator i = thms.begin(), iend = thms.end();
        i != iend; ++i) {
      // Check that t is c == d or c IFF d
      if(CHECK_PROOFS)
        CHECK_SOUND(i->isRewrite(),
                    "CVCL::CommonTheoremProducer::substitutivityRule:\n  "
                    "premis is not an equality or IFF: "
                    + i->getExpr().toString());
                    // + "\n  op = " + ((Op) e.getOp).toString());
      pfs.push_back(i->getProof());
      pf = newPf("optimized_subst_op",e,e2,pfs);
  }
  Theorem res = newRWTheorem(e, e2, a, pf);
  IF_DEBUG(debugger.setElapsed(tmpTimer);
	   accum0 += tmpTimer);
  return res;
}

////////////////////////////////////////////////////////////////////////
// TCC rules (3-valued logic)
////////////////////////////////////////////////////////////////////////

//  G1 |- phi   G2 |- D_phi
// -------------------------
//       G1,G2 |-_3 phi
Theorem3
CommonTheoremProducer::queryTCC(const Theorem& phi, const Theorem& D_phi) {
  Assumptions a;
  Proof pf;
  if(CHECK_PROOFS)
    CHECK_SOUND(D_phi.getExpr() == d_core->getTCC(phi.getExpr()),
		"CommonTheoremProducer::queryTCC: "
		"bad TCC for a formula:\n\n  "
		+phi.getExpr().toString()
		+"\n\n  TCC must be the following:\n\n  "
		+d_core->getTCC(phi.getExpr()).toString()
		+"\n\nBut given this as a TCC:\n\n  "
		+D_phi.getExpr().toString());
  if(withAssumptions())
    a = merge(phi.getAssumptions(), D_phi);
  if(withProof()) {
    vector<Expr> args;
    vector<Proof> pfs;
    args.push_back(phi.getExpr());
    args.push_back(D_phi.getExpr());
    pfs.push_back(phi.getProof());
    pfs.push_back(D_phi.getProof());
    pf = newPf("queryTCC", args, pfs);
    }
  return newTheorem3(phi.getExpr(), a, pf);
}


//  G0,a1,...,an |-_3 phi  G1 |- D_a1 ... Gn |- D_an
// -------------------------------------------------
//    G0,G1,...,Gn |-_3 (a1 & ... & an) -> phi
Theorem3
CommonTheoremProducer::implIntro3(const Theorem3& phi,
				  const std::vector<Expr>& assump,
				  const vector<Theorem>& tccs) {
  bool checkProofs(CHECK_PROOFS);
  // This rule only makes sense when runnnig with assumptions
  if(checkProofs) {
    CHECK_SOUND(withAssumptions(),
		"implIntro3: called while running without assumptions");
  }

  const Assumptions& phiAssump = phi.getAssumptionsRef();

  if(checkProofs) {
    CHECK_SOUND(assump.size() == tccs.size(),
		"implIntro3: number of assumptions ("
		+int2string(assump.size())+") and number of TCCs ("
		+int2string(tccs.size())
		+") does not match");
    for(size_t i=0; i<assump.size(); i++) {
      const Theorem& thm = phiAssump[assump[i]];
      CHECK_SOUND(!thm.isNull() && thm.isAxiom(),
		  "implIntro3: this is not an assumption of phi:\n\n"
		  "  a["+int2string(i)+"] = "+assump[i].toString()
		  +"\n\n  phi = "+phi.getExpr().toString());
      CHECK_SOUND(tccs[i].getExpr() == d_core->getTCC(assump[i]),
		  "implIntro3: Assumption does not match its TCC:\n\n"
		  "  a["+int2string(i)+"] = "+assump[i].toString()
		  +"  TCC(a["+int2string(i)+"]) = "
		  +d_core->getTCC(assump[i]).toString()
		  +"\n\n  tccs["+int2string(i)+"] = "
		  +tccs[i].getExpr().toString());
    }
  }

  // Proof compaction: trivial derivation
  if(assump.size() == 0) return phi;

  Assumptions a((phiAssump - assump) + merge(tccs));
  Proof pf;
  if(withProof()) {
    vector<Proof> u; // Proof labels for assumptions
    for(vector<Expr>::const_iterator i=assump.begin(), iend=assump.end();
	i!=iend; ++i) {
      const Theorem& t = phiAssump[*i];
      u.push_back(t.getProof());
    }
    // Arguments to the proof rule:
    // impl_intro_3(phi, a1,...,an,tcc1,...tccn,pf_tcc1,...pf_tccn,
    //              [lambda(a1,...,an): pf_phi])
    vector<Expr> args;
    vector<Proof> pfs;
    args.push_back(phi.getExpr());
    args.insert(args.end(), assump.begin(), assump.end());
    for(vector<Theorem>::const_iterator i=tccs.begin(), iend=tccs.end();
	i!=iend; ++i) {
      args.push_back(i->getExpr());
      pfs.push_back(i->getProof());
    }
    // Lambda-abstraction of pf_phi
    pfs.push_back(newPf(u, assump, phi.getProof()));
    pf = newPf("impl_intro_3", args, pfs);
  }
  Expr conj(andExpr(assump));
  return newTheorem3(conj.impExpr(phi.getExpr()), a, pf);
}


// |- e,  |- !e ==> |- FALSE
Theorem
CommonTheoremProducer::contradictionRule(const Theorem& e,
					const Theorem& not_e) {
  Assumptions a;
  Proof pf;
  if(CHECK_PROOFS)
    CHECK_SOUND(!e.getExpr() == not_e.getExpr(),
		"CommonTheoremProducer::contraditionRule: "
		"theorems don't match:\n e = "+e.toString()
		+"\n not_e = "+not_e.toString());
  if(withAssumptions()) a = merge(e, not_e);
  if(withProof()) {
    vector<Proof> pfs;
    pfs.push_back(e.getProof());
    pfs.push_back(not_e.getProof());
    pf = newPf("contradition", e.getExpr(), pfs);
  }
  return newTheorem(d_em->falseExpr(), a, pf);
}

// e ==> e IFF TRUE
Theorem
CommonTheoremProducer::iffTrue(const Theorem& e)
{
  Assumptions a;
  Proof pf;
  if (withAssumptions()) a = e.getAssumptions().copy();
  if(withProof()) {
    pf = newPf("iff_true", e.getExpr(), e.getProof());
  }
  return newRWTheorem(e.getExpr(), d_em->trueExpr(), a, pf);
}

// e ==> !e IFF FALSE
Theorem
CommonTheoremProducer::iffNotFalse(const Theorem& e) {
  Assumptions a;
  Proof pf;
  if (withAssumptions()) a = e.getAssumptions().copy();
  if(withProof()) {
    pf = newPf("iff_not_false", e.getExpr(), e.getProof());
  }
  return newRWTheorem(!e.getExpr(), d_em->falseExpr(), a, pf);
}

// e IFF TRUE ==> e
Theorem
CommonTheoremProducer::iffTrueElim(const Theorem& e) {
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isRewrite() && e.getRHS().isTrue(),
		"CommonTheoremProducer::iffTrueElim: "
		"theorem is not e<=>TRUE: "+ e.toString());
  Assumptions a;
  Proof pf;
  if (withAssumptions()) a = e.getAssumptions().copy();
  if(withProof()) {
    pf = newPf("iff_true_elim", e.getLHS(), e.getProof());
  }
  return newTheorem(e.getLHS(), a, pf);
}


// e IFF FALSE ==> !e
Theorem
CommonTheoremProducer::iffFalseElim(const Theorem& e) {
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isRewrite() && e.getRHS().isFalse(),
		"CommonTheoremProducer::iffFalseElim: "
		"theorem is not e<=>FALSE: "+ e.toString());
  const Expr& lhs = e.getLHS();
  Assumptions a;
  Proof pf;
  if (withAssumptions()) a = e.getAssumptions().copy();
  if(withProof()) {
    pf = newPf("iff_false_elim", lhs, e.getProof());
  }
  return newTheorem(!lhs, a, pf);
}

// e1 <=> e2  ==>  ~e1 <=> ~e2
Theorem
CommonTheoremProducer::iffContrapositive(const Theorem& e) {
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isRewrite() && getType(e.getRHS()).isBool(),
		"CommonTheoremProducer::iffContrapositive: "
		"theorem is not e1<=>e2: "+ e.toString());
  Assumptions a;
  Proof pf;
  if (withAssumptions()) a = e.getAssumptions().copy();
  if(withProof()) {
    pf = newPf("iff_contrapositive", e.getExpr(), e.getProof());
  }
  return newRWTheorem(e.getLHS().negate(),e.getRHS().negate(), a, pf);
}


// !!e ==> e
Theorem
CommonTheoremProducer::notNotElim(const Theorem& not_not_e) {
  if(CHECK_PROOFS)
    CHECK_SOUND(not_not_e.getExpr().isNot() && not_not_e.getExpr()[0].isNot(),
		"CommonTheoremProducer::notNotElim: bad theorem: !!e = "
		+ not_not_e.toString());
  Assumptions a;
  Proof pf;
  if(withAssumptions()) a = not_not_e.getAssumptionsCopy();
  if(withProof())
    pf = newPf("not_not_elim", not_not_e.getExpr(), not_not_e.getProof());
  return newTheorem(not_not_e.getExpr()[0][0], a, pf);
}

// e1 AND (e1 IFF e2) ==> e2
Theorem
CommonTheoremProducer::iffMP(const Theorem& e1, const Theorem& e1_iff_e2) {
  if(CHECK_PROOFS) {
    CHECK_SOUND(e1_iff_e2.isRewrite(),
		"iffMP: not IFF: "+e1_iff_e2.toString());
    CHECK_SOUND(e1.getExpr() == (e1_iff_e2.getLHS()),
		"iffMP: theorems don't match:\n  e1 = " + e1.toString()
		+ ", e1_iff_e2 = " + e1_iff_e2.toString());
  }
  const Expr& e2(e1_iff_e2.getRHS());
  // Avoid e1.getExpr(), it may cause unneeded Expr creation
  if (e1_iff_e2.getLHS() == e2) return e1;
  Assumptions a;
  Proof pf;
  if(withAssumptions()) a = merge(e1, e1_iff_e2);
  if(withProof()) {
    vector<Proof> pfs;
    pfs.push_back(e1.getProof());
    pfs.push_back(e1_iff_e2.getProof());
    pf = newPf("iff_mp", e1.getExpr(), e2, pfs);
  }
  return newTheorem(e2, a, pf);
}


// e1 AND (e1 IMPLIES e2) ==> e2
Theorem
CommonTheoremProducer::implMP(const Theorem& e1, const Theorem& e1_impl_e2) {
  const Expr& impl = e1_impl_e2.getExpr();
  if(CHECK_PROOFS) {
    CHECK_SOUND(impl.isImpl() && impl.arity()==2,
		"implMP: not IMPLIES: "+impl.toString());
    CHECK_SOUND(e1.getExpr() == impl[0],
		"implMP: theorems don't match:\n  e1 = " + e1.toString()
		+ ", e1_impl_e2 = " + impl.toString());
  }
  const Expr& e2 = impl[1];
  // Avoid e1.getExpr(), it may cause unneeded Expr creation
  if (impl[0] == e2) return e1;
  Assumptions a;
  Proof pf;
  if(withAssumptions()) a = merge(e1, e1_impl_e2);
  if(withProof()) {
    vector<Proof> pfs;
    pfs.push_back(e1.getProof());
    pfs.push_back(e1_impl_e2.getProof());
    pf = newPf("impl_mp", e1.getExpr(), e2, pfs);
  }
  return newTheorem(e2, a, pf);
}


// AND(e_0,...e_{n-1}) ==> e_i
Theorem
CommonTheoremProducer::andElim(const Theorem& e, int i) {
  if(CHECK_PROOFS) {
    CHECK_SOUND(e.getExpr().isAnd(), "andElim: not an AND: " + e.toString());
    CHECK_SOUND(i < e.getExpr().arity(), "andElim: i = " + int2string(i)
		+ " >= arity = " + int2string(e.getExpr().arity())
		+ " in " + e.toString());
  }
  Assumptions a;
  Proof pf;
  if(withAssumptions())
    a = e.getAssumptions().copy();
  if(withProof())
    pf = newPf("andE", newRatExpr(d_em, i), e.getExpr(), e.getProof());
  return newTheorem(e.getExpr()[i], a, pf);
}

//! e1, e2 ==> AND(e1, e2)
Theorem
CommonTheoremProducer::andIntro(const Theorem& e1, const Theorem& e2) {
  vector<Theorem> thms;
  thms.push_back(e1);
  thms.push_back(e2);
  return andIntro(thms);
}

Theorem
CommonTheoremProducer::andIntro(const vector<Theorem>& es) {
  Assumptions a;
  Proof pf;
  if(CHECK_PROOFS)
    CHECK_SOUND(es.size() > 0,
		"andIntro(vector<Theorem>): vector must be non-empty");
  if(withAssumptions()) {
    a = merge(es);
  }
  if(withProof()) {
    vector<Proof> pfs;
    for(vector<Theorem>::const_iterator i=es.begin(), iend=es.end();
	i!=iend; ++i)
      pfs.push_back(i->getProof());
    pf = newPf("andI", pfs);
  }
  vector<Expr> kids;
  for(vector<Theorem>::const_iterator i=es.begin(), iend=es.end();
      i!=iend; ++i)
    kids.push_back(i->getExpr());
  return newTheorem(andExpr(kids), a, pf);
}

// e: T ==> |- typePred_T(e)  [asserting the type predicate of e]
Theorem
CommonTheoremProducer::typePred(const Expr& e) {
  Type tp(getType(e));
  Expr pred(d_core->getTypePred(tp, e));
  Assumptions a;
  Proof pf;
  if(withProof()) {
    pf = newPf("type_pred", e, tp.getExpr());
  }
  return newTheorem(pred, a, pf);
}


//  G,a1,...,an |- phi
// -------------------------------------------------
//  G |- (a1 & ... & an) -> phi
Theorem
CommonTheoremProducer::implIntro(const Theorem& phi,
				 const std::vector<Expr>& assump) {
  bool checkProofs(CHECK_PROOFS);
  // This rule only makes sense when runnnig with assumptions
  if(checkProofs) {
    CHECK_SOUND(withAssumptions(),
		"implIntro: called while running without assumptions");
  }

  const Assumptions& phiAssump = phi.getAssumptionsRef();

  if(checkProofs) {
    for(size_t i=0; i<assump.size(); i++) {
      const Theorem& thm = phiAssump[assump[i]];
      CHECK_SOUND(!thm.isNull() && thm.isAxiom(),
		  "implIntro: this is not an assumption of phi:\n\n"
		  "  a["+int2string(i)+"] = "+assump[i].toString()
		  +"\n\n  phi = "+phi.getExpr().toString());
    }
  }

  // Proof compaction: trivial derivation
  if(assump.size() == 0) return phi;

  Assumptions a(phiAssump - assump);
  Proof pf;
  if(withProof()) {
    vector<Proof> u; // Proof labels for assumptions
    for(vector<Expr>::const_iterator i=assump.begin(), iend=assump.end();
	i!=iend; ++i) {
      const Theorem& t = phiAssump[*i];
      u.push_back(t.getProof());
    }
    // Arguments to the proof rule:
    // impl_intro_3(phi, a1,...,an,tcc1,...tccn,pf_tcc1,...pf_tccn,
    //              [lambda(a1,...,an): pf_phi])
    vector<Expr> args;
    vector<Proof> pfs;
    args.push_back(phi.getExpr());
    args.insert(args.end(), assump.begin(), assump.end());
    // Lambda-abstraction of pf_phi
    pfs.push_back(newPf(u, assump, phi.getProof()));
    pf = newPf("impl_intro", args, pfs);
  }
  Expr conj(andExpr(assump));
  return newTheorem(conj.impExpr(phi.getExpr()), a, pf);
}


// e1 => e2  ==>  ~e2 => ~e1
Theorem
CommonTheoremProducer::implContrapositive(const Theorem& thm) {
  const Expr& impl = thm.getExpr();
  if(CHECK_PROOFS) {
    CHECK_SOUND(impl.isImpl() && impl.arity()==2,
		"CommonTheoremProducer::implContrapositive: thm="
		+impl.toString());
  }
  Assumptions a;
  Proof pf;
  if(withAssumptions()) a = thm.getAssumptionsCopy();
  if(withProof())
    pf = newPf("impl_contrapositive", impl, thm.getProof());
  return newTheorem(impl[1].negate().impExpr(impl[0].negate()), a, pf);
}


// NOT e ==> e IFF FALSE
Theorem
CommonTheoremProducer::notToIff(const Theorem& not_e)
{
  if(CHECK_PROOFS)
    CHECK_SOUND(not_e.getExpr().isNot(),
		"notToIff: not NOT: "+not_e.toString());

  // Make it an atomic rule (more efficient)
  Expr e(not_e.getExpr()[0]);
  Assumptions a;
  Proof pf;
  if(withAssumptions()) a = not_e.getAssumptions().copy();
  if(withProof())
    pf=newPf("not_to_iff", e, not_e.getProof());
  return newRWTheorem(e, d_em->falseExpr(), a, pf);
}

// Rewrite rules used only in TheoryCore to replace LETDECL and
// CONSTDEF with their definitions
Theorem
CommonTheoremProducer::rewriteLetDecl(const Expr& e) {
  if(CHECK_PROOFS)
    CHECK_SOUND(e.getKind() == LETDECL,
		"rewriteLetDecl: wrong expression: " + e.toString());
  Assumptions a;
  Proof pf;
  if(withProof()) // FIXME: implement this in flea
    pf = newPf("rewrite_letdecl", e[1]);
  return newRWTheorem(e, e[1], a, pf);
}

Theorem
CommonTheoremProducer::rewriteConstDef(const Expr& e) {
  if(CHECK_PROOFS)
    CHECK_SOUND(e.getKind() == CONSTDEF,
		"rewriteConstDef: wrong expression: " + e.toString());
  Assumptions a;
  Proof pf;
  if(withProof()) // FIXME: implement this in flea
    pf = newPf("rewrite_constdef", e[1]);
  return newRWTheorem(e, e[1], a, pf);
}

// ==> (a == a) IFF TRUE
Theorem
CommonTheoremProducer::rewriteReflexivity(const Expr& t) {
  if(CHECK_PROOFS)
    CHECK_SOUND((t.isEq() || t.isIff()) && t[0] == t[1],
		"rewriteReflexivity: wrong expression: "
		+ t.toString());
  Assumptions a;
  Proof pf;
  if(withProof()) {
    if(t.isEq()) {
      DebugAssert(!getType(t[0]).isNull(),
		  "rewriteReflexivity: t[0] has no type: "
		  + t[0].toString());
      pf = newPf("rewrite_eq_refl", getType(t[0]).getExpr(), t[0]);
    } else
      pf = newPf("rewrite_iff_refl", t[0]);
  }
  return newRWTheorem(t, d_em->trueExpr(), a, pf);
}

// ==> NOT TRUE IFF FALSE
Theorem
CommonTheoremProducer::rewriteNotTrue(const Expr& e) {
  Assumptions a;
  Proof pf;
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isNot() && e[0].isTrue(),
		"rewriteNotTrue precondition violated");
  if(withProof())
    pf = newPf("rewrite_not_true");
  return newRWTheorem(e, d_em->falseExpr(), a, pf);
}

// ==> NOT FALSE IFF TRUE
Theorem
CommonTheoremProducer::rewriteNotFalse(const Expr& e) {
  Assumptions a;
  Proof pf;
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isNot() && e[0].isFalse(),
		"rewriteNotFalse precondition violated");
  if(withProof())
    pf = newPf("rewrite_not_false");
  return newRWTheorem(e, d_em->trueExpr(), a, pf);
}
  
// ==> (NOT NOT e) IFF e, takes !!e
Theorem
CommonTheoremProducer::rewriteNotNot(const Expr& e) {
  Assumptions a;
  Proof pf;
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isNot() && e[0].isNot(),
		"rewriteNotNot precondition violated");
  if(withProof())
    pf = newPf("rewrite_not_not", e[0][0]);
  return newRWTheorem(e, e[0][0], a, pf);
}

// ==> NOT (AND e1 ... en) IFF (OR !e1 ... !en), takes (AND ...)
Theorem
CommonTheoremProducer::rewriteNotAnd(const Expr& e) {
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isNot() && e[0].isAnd(),
		"rewriteNotAnd: precondition violated: " + e.toString());
  vector<Expr> kids; // vector of <!e1,...,!en>
  for(Expr::iterator i=e[0].begin(), iend=e[0].end(); i!=iend; ++i)
    // Collapse double negations
    kids.push_back(i->negate());
  Assumptions a;
  Proof pf;
  if(withProof())
    pf = newPf("rewrite_not_and", e[0]);
  return newRWTheorem(e, orExpr(kids), a, pf);
}

// ==> NOT (OR e1 ... en) IFF (AND !e1 ... !en), takes (OR ...)
Theorem
CommonTheoremProducer::rewriteNotOr(const Expr& e) {
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isNot() && e[0].isOr(),
		"rewriteNotOr: precondition violated: " + e.toString());
  vector<Expr> kids; // vector of <!e1,...,!en>
  for(Expr::iterator i=e[0].begin(), iend=e[0].end(); i!=iend; ++i)
    // Collapse double negations
    kids.push_back(i->negate());
  Assumptions a;
  Proof pf;
  if(withProof())
    pf = newPf("rewrite_not_or", e);
  return newRWTheorem(e, andExpr(kids), a, pf);
}

// ==> NOT IFF(e1,e2) IFF IFF(e1,NOT e2)
Theorem
CommonTheoremProducer::rewriteNotIff(const Expr& e) {
  Assumptions a;
  Proof pf;
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isNot() && e[0].isIff(), "rewriteNotIff precondition violated");
  if(withProof()) {
    pf = newPf("rewrite_not_iff", e);
  }
  return newRWTheorem(e, e[0][0].iffExpr(!e[0][1]), a, pf);
}

// ==> NOT ITE(a,b,c) IFF ITE(a,NOT b,NOT c)
Theorem
CommonTheoremProducer::rewriteNotIte(const Expr& e) {
  Assumptions a;
  Proof pf;
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isNot() && e[0].isITE(), "rewriteNotIte precondition violated");
  if(withProof()) {
    pf = newPf("rewrite_not_ite", e);
  }
  return newRWTheorem(e, e[0][0].iteExpr(!e[0][1], !e[0][2]), a, pf);
}

//! ==> NOT FORALL (vars): e  IFF EXISTS (vars) NOT e
Theorem
CommonTheoremProducer::rewriteNotForall(const Expr& e) {
  if(CHECK_PROOFS) {
    CHECK_SOUND(e.isNot() && e[0].isForall(),
		"rewriteNotForall: expr must be NOT FORALL:\n"
		+e.toString());
  }
  Assumptions a;
  Proof pf;
  if(withProof())
    pf = newPf("rewrite_not_forall", e);
  return newRWTheorem(e, newClosureExpr(EXISTS, e[0].getVars(),
					!e[0].getQuantBody()), a, pf);
}


//! ==> NOT EXISTS (vars): e  IFF FORALL (vars) NOT e
Theorem
CommonTheoremProducer::rewriteNotExists(const Expr& e) {
  if(CHECK_PROOFS) {
    CHECK_SOUND(e.isNot() && e[0].isExists(),
		"rewriteNotExists: expr must be NOT FORALL:\n"
		+e.toString());
  }
  Assumptions a;
  Proof pf;
  if(withProof())
    pf = newPf("rewrite_not_exists", e);
  return newRWTheorem(e, newClosureExpr(FORALL, e[0].getVars(),
					!e[0].getQuantBody()), a, pf);
}


// ==> ITE(TRUE, e1, e2) == e1
Theorem
CommonTheoremProducer::rewriteIteTrue(const Expr& e) {
  Assumptions a;
  Proof pf;
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isITE() && e[0].isTrue(),
		"rewriteIteTrue precondition violated");
  if(withProof()) {
    Type t = getType(e[1]);
    DebugAssert(!t.isNull(), "rewriteIteTrue: e1 has no type: "
		+ e[1].toString());
    bool useIff = t == d_tm->getVCL()->boolType();
    if(useIff)
      pf = newPf("rewrite_ite_true_iff", e[1], e[2]);
    else {
      pf = newPf("rewrite_ite_true", t.getExpr(), e[1], e[2]);
    }
  }
  return newRWTheorem(e, e[1], a, pf);
}

// ==> ITE(FALSE, e1, e2) == e2
Theorem
CommonTheoremProducer::rewriteIteFalse(const Expr& e) {
  Assumptions a;
  Proof pf;
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isITE() && e[0].isFalse(),
		"rewriteIteFalse precondition violated");
  if(withProof()) {
    Type t = getType(e[1]);
    DebugAssert(!t.isNull(), "rewriteIteFalse: e1 has no type: "
		+ e[1].toString());
    bool useIff = t == d_tm->getVCL()->boolType();
    if(useIff)
      pf = newPf("rewrite_ite_false_iff", e[1], e[2]);
    else {
      pf = newPf("rewrite_ite_false", t.getExpr(), e[1], e[2]);
    }
  }
  return newRWTheorem(e, e[2], a, pf);
}

// ==> ITE(c, e, e) == e
Theorem
CommonTheoremProducer::rewriteIteSame(const Expr& e) {
  Assumptions a;
  Proof pf;
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isITE() && e[1] == e[2],
		"rewriteIteSame precondition violated");
  if(withProof()) {
    Type t = getType(e[1]);
    DebugAssert(!t.isNull(), "rewriteIteSame: e[1] has no type: "
		+ e[1].toString());
    bool useIff = t == d_tm->getVCL()->boolType();
    if(useIff)
      pf = newPf("rewrite_ite_same_iff", e[0], e[1]);
    else {
      pf = newPf("rewrite_ite_same", t.getExpr(), e[0], e[1]);
    }
  }
  return newRWTheorem(e, e[1], a, pf);
}

// a |- b == d ==> ITE(a, b, c) == ITE(a, d, c)
Theorem
CommonTheoremProducer::rewriteIteThen(const Expr& e, const Theorem& thenThm) {
  Assumptions a;
  Proof pf;
  if(CHECK_PROOFS) {
    CHECK_SOUND(withAssumptions(), "Cannot check proof without assumptions");
    CHECK_SOUND(e.isITE() && thenThm.isRewrite() && e[1] == thenThm.getLHS(),
		"rewriteIteThen precondition violated \n then expression: "
		+ thenThm.getExpr().toString() + "\n e = " + e.toString());
  }
  if (withAssumptions()) {
    a = thenThm.getAssumptions() - e[0];
  }
  if(withProof()) {
    Type t = getType(e);
    DebugAssert(!t.isNull(), "rewriteIteThen: e has no type: "
		+ e.toString());
    bool useIff = t == d_tm->getVCL()->boolType();
    if(useIff)
      pf = newPf("rewrite_ite_then_iff", e, thenThm.getProof());
    else {
      pf = newPf("rewrite_ite_then", e, thenThm.getProof());
    }
  }
  return newRWTheorem(e, e[0].iteExpr(thenThm.getRHS(), e[2]), a, pf);
}

// !a |- c == d ==> ITE(a, b, c) == ITE(a, b, d)
Theorem
CommonTheoremProducer::rewriteIteElse(const Expr& e, const Theorem& elseThm) {
  Assumptions a;
  Proof pf;
  if(CHECK_PROOFS) {
    CHECK_SOUND(withAssumptions(), "Cannot check proof without assumptions");
    CHECK_SOUND(e.isITE() && elseThm.isRewrite() && e[2] == elseThm.getLHS(),
		"rewriteIteElse precondition violated \n else expression: "
		+ elseThm.getExpr().toString() + "\n e = " + e.toString());
  }
  if (withAssumptions()) {
    a = elseThm.getAssumptions() - !e[0];
  }
  if(withProof()) {
    Type t = getType(e);
    DebugAssert(!t.isNull(), "rewriteIteElse: e has no type: "
		+ e.toString());
    bool useIff = t == d_tm->getVCL()->boolType();
    if(useIff)
      pf = newPf("rewrite_ite_else_iff", e, elseThm.getProof());
    else {
      pf = newPf("rewrite_ite_else", e, elseThm.getProof());
    }
  }
  return newRWTheorem(e, e[0].iteExpr(e[1], elseThm.getRHS()), a, pf);
}

// ==> ITE(c, e1, e2) <=> (!c OR e1) AND (c OR e2)
Theorem 
CommonTheoremProducer::rewriteIteBool(const Expr& c,
				      const Expr& e1, const Expr& e2) {
  if(CHECK_PROOFS)
    CHECK_SOUND(getType(e1).isBool() && getType(e2).isBool(),
		"rewriteIteBool: Not a boolean ITE: "
		+ c.iteExpr(e1, e2).toString());
  Assumptions a;
  Proof pf;
  if(withProof())
    pf = newPf("rewrite_ite_bool", c, e1, e2);
  return newRWTheorem(c.iteExpr(e1,e2),
		      (e1.orExpr(!c).andExpr(c.orExpr(e2))), a, pf);
}


// ==> AND(e_1,...,e_n) IFF <simplified expr>
// 1) if e_i = FALSE then return FALSE
// 2) if e_i = TRUE, remove it from children
// 3) if e_i = AND(f_1,...,f_m) then AND(e_1,...,e_{i-1},f_1,...,f_m,e_{i+1},...,e_n)
// 4) if n=0 return TRUE
// 5) if n=1 return e_1
Theorem
CommonTheoremProducer::rewriteAnd(const Expr& e) {
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isAnd(), "rewriteAnd: bad Expr: " + e.toString());
  Assumptions a;
  Proof pf;
  const vector<Expr>& kids = e.getKids();
  unsigned k(0), ar(kids.size());
  ExprMap<bool> newKids;
  bool isFalse (false);
  bool changed(false);
  for(; k < ar; ++k) {
    if (kids[k].isFalse()) { isFalse=true; break; }
    if (kids[k].isAnd() || kids[k].isTrue()) {
      unsigned j;
      if (!changed) {
        changed = true;
        for (j = 0; j < k; ++j) newKids[kids[j]] = true;
      }
      if (kids[k].isAnd()) {
        unsigned ar2(kids[k].arity());
        for (j = 0; j < ar2; ++j) newKids[kids[k][j]] = true;
      }
    }
    else if (changed) newKids[kids[k]]=true;
  }
  Expr res;
  if (isFalse) res = d_em->falseExpr();
  else if (!changed) {
    switch(ar) {
    case 0: res = d_em->trueExpr(); break;
    case 1: res = e[0]; break;
    default: res = e;
    }
  }
  else if (newKids.size() == 0) res = d_em->trueExpr(); // All newKids were TRUE
  else if (newKids.size() == 1)
    res = newKids.begin()->first; // The only child
  else {
    vector<Expr> v;
    for(ExprMap<bool>::iterator i=newKids.begin(), iend=newKids.end();
        i!=iend; ++i)
      v.push_back(i->first);
    res = andExpr(v);
  }
  if(withProof()) {
    pf = newPf("rewrite_and", e);
  }
  return newRWTheorem(e, res, a, pf);
}

// ==> OR(e1,e2) IFF <simplified expr>
Theorem 
CommonTheoremProducer::rewriteOr(const Expr& e) {
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isOr(), "rewriteOr: bad Expr: " + e.toString());
  Assumptions a;
  Proof pf;
  const vector<Expr>& kids = e.getKids();
  unsigned k(0), ar(kids.size());
  ExprMap<bool> newKids;
  bool isTrue (false);
  bool changed(false);
  for (; k < ar; ++k) {
    if (kids[k].isTrue()) { isTrue=true; break; }
    if (kids[k].isOr() || kids[k].isFalse()) {
      unsigned j;
      if (!changed) {
	changed = true;
	for (j = 0; j < k; ++j) newKids[kids[j]]=true;
      }
      if (kids[k].isOr()) {
	unsigned ar2(kids[k].arity());
	for (j = 0; j < ar2; ++j) newKids[kids[k][j]]=true;
      }
    }
    else if (changed) newKids[kids[k]];
  }
  Expr res;
  if (isTrue) res = d_em->trueExpr();
  else if (!changed) {
    switch(ar) {
    case 0: res = d_em->falseExpr(); break;
    case 1: res = e[0]; break;
    default: res = e;
    }
  }
  else if (newKids.size() == 0) res = d_em->falseExpr(); // All kids were FALSE
  else if (newKids.size() == 1) res = newKids.begin()->first; // The only child
  else {
    vector<Expr> v;
    for(ExprMap<bool>::iterator i=newKids.begin(), iend=newKids.end();
        i!=iend; ++i)
      v.push_back(i->first);
    res = orExpr(v);
  }
  if(withProof()) {
    pf = newPf("rewrite_or", e);
  }
  return newRWTheorem(e, res, a, pf);
}

// ==> IMPLIES(e1,e2) IFF OR(!e1, e2)
Theorem
CommonTheoremProducer::rewriteImplies(const Expr& e) {
  Assumptions a;
  Proof pf;
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isImpl(), "rewriteImplies precondition violated");
  if(withProof()) {
    pf = newPf("rewrite_implies", e[0], e[1]);
  }
  return newRWTheorem(e, !e[0] || e[1], a, pf);
}

// ==> IFF(e1,e2) IFF <simplified expr>
Theorem
CommonTheoremProducer::rewriteIff(const Expr& e) {
  Assumptions a;
  Proof pf;
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isIff(), "rewriteIff precondition violated");
  if(withProof()) {
    pf = newPf("rewrite_iff", e[0], e[1]);
  }

  if(e[0] == e[1]) return rewriteReflexivity(e);

  switch(e[0].getKind()) {
  case TRUE:
    return newRWTheorem(e, e[1], a, pf);
  case FALSE:
    return newRWTheorem(e, !e[1], a ,pf);
  case NOT:
    if(e[0][0]==e[1]) 
      return newRWTheorem(e, d_em->falseExpr(), a, pf);
    break;
  default: break;
  }
  
  switch(e[1].getKind()) {
  case TRUE:
    return newRWTheorem(e, e[0], a, pf);
  case FALSE:
    return newRWTheorem(e, !e[0], a, pf);
  case NOT:
    if(e[0]==e[1][0]) 
      return newRWTheorem(e, d_em->falseExpr(), a, pf);
    break;
  default:
    break;
  }

  if(e[0] < e[1])
    return rewriteUsingSymmetry(e);
  else
    return reflexivityRule(e);
}

// ==> IFF(e1,e2) IFF AND(OR(!e1, e2),OR(e1, !e2))
Theorem
CommonTheoremProducer::iffToAndOr(const Expr& e) {
  Assumptions a;
  Proof pf;
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isIff(), "rewriteIff precondition violated");
  if(withProof()) {
    pf = newPf("iffToAndOr", e[0], e[1]);
  }
  if(e[0] == e[1]) return rewriteReflexivity(e);
  if (e[0].isTrue()) return newRWTheorem(e, e[1], a, pf);
  else if (e[1].isTrue()) return newRWTheorem(e, e[0], a, pf);
  else if (e[0].isFalse()) return newRWTheorem(e, !e[1], a ,pf);
  else if (e[1].isFalse()) return newRWTheorem(e, !e[0], a, pf);
  Expr impl1(e[1] || !e[0]), impl2(e[0] || !e[1]);
  return newRWTheorem(e, impl1 && impl2, a, pf);
}


// ==> NOT(e) == ITE(e, FALSE, TRUE)
Theorem
CommonTheoremProducer::NotToIte(const Expr& not_e){
  Assumptions a;
  Proof pf;
  if(CHECK_PROOFS)
    CHECK_SOUND(not_e.isNot() && getType(not_e[0]).isBool(),
		"NotToIte precondition violated");
  if(withProof())
    pf = newPf("NotToIte", not_e[0]);
  if(not_e[0].isTrue())
    return rewriteNotTrue(not_e);
  else if(not_e[0].isFalse())
    return rewriteNotFalse(not_e);
  Expr ite(not_e[0].iteExpr(d_em->falseExpr(), d_em->trueExpr()));
  return newRWTheorem(not_e, ite, a, pf);  
}

// ==> Or(e) == ITE(e[1], TRUE, e[0])
Theorem
CommonTheoremProducer::OrToIte(const Expr& e){
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isOr(),
		"OrToIte: precondition violated: " + e.toString());
  Assumptions a;
  Proof pf;
  if(withProof()) {
    pf = newPf("OrToIte", e);
  }
  const vector<Expr>& kids = e.getKids();
  unsigned k(kids.size());
  if(k==0)
    return newRWTheorem(e, d_em->falseExpr(), a, pf);
  if(k==1)
    return newRWTheorem(e, e[0], a, pf);
  Expr ite(kids[k-1]);
  if(CHECK_PROOFS)
    CHECK_SOUND(!getType(ite).isNull(),
		"OrToIte: kid " + int2string(k-1) + " has no type: "
		+ (ite).toString());
  for (; k > 1; k--) {
    if (kids[k-2].isTrue()) {
      ite = d_em->trueExpr();
      break;
    }
    else if(kids[k-2].isFalse()) continue;
    else{
      if(CHECK_PROOFS)
	CHECK_SOUND(!getType(kids[k-2]).isNull(),
		  "OrToIte: kid " + int2string(k-2) + " has no type: "
		  + (kids[k-2]).toString());
      ite = ite.iteExpr(d_em->trueExpr(), kids[k-2]);
    }
  }
  return newRWTheorem(e, ite, a, pf);
}

// ==> And(e) == ITE(e[1], e[0], FALSE)
Theorem
CommonTheoremProducer::AndToIte(const Expr& e){
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isAnd(),
		"AndToIte: precondition violated: " + e.toString());
  Assumptions a;
  Proof pf;
  if(withProof()) {
    pf = newPf("AndToIte", e);
  }
  const vector<Expr>& kids = e.getKids();
  unsigned k(kids.size());
  if(k==0)
    return newRWTheorem(e, d_em->trueExpr(), a, pf);
  if(k==1)
    return newRWTheorem(e, e[0], a, pf);
  Expr ite(kids[k-1]);
  if(CHECK_PROOFS)
    CHECK_SOUND(!getType(ite).isNull(),
		"AndToIte: kid " + int2string(k-1) + " has no type: "
		+ (ite).toString());
  for (; k > 1; k--) {
    if (kids[k-2].isFalse()) {
      ite = d_em->falseExpr();
      break;
    }
    else if(kids[k-2].isTrue()) {
      continue;
    }
    else{
      if(CHECK_PROOFS)
	CHECK_SOUND(!getType(kids[k-2]).isNull(),
		    "AndToIte: kid " + int2string(k-2) + " has no type: "
		    + (kids[k-2]).toString());
      ite = ite.iteExpr(kids[k-2], d_em->falseExpr());
    }
  }
  return newRWTheorem(e, ite, a, pf);
}

// ==> IFF(a,b) == ITE(a, b, !b)
Theorem
CommonTheoremProducer::IffToIte(const Expr& e){
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isIff() && getType(e[0]).isBool() && getType(e[1]).isBool(),
		"IffToIte: precondition violated: " + e.toString());
  Assumptions a;
  Proof pf;
  if(e[0] == e[1]) return reflexivityRule(e);
  Expr ite(e[0].iteExpr(e[1], e[1].iteExpr(d_em->falseExpr(),
					   d_em->trueExpr())));
  if(withProof()) {
    pf = newPf("IffToIte", e);
  }
  return newRWTheorem(e, ite, a, pf); 
}

// ==> IMPLIES(a,b) == ITE(a, b, TRUE)
Theorem
CommonTheoremProducer::ImpToIte(const Expr& e){
  if(CHECK_PROOFS)
    CHECK_SOUND(e.isImpl() && getType(e[0]).isBool() && getType(e[1]).isBool(),
		"ImpToIte: precondition violated: " + e.toString());
  Assumptions a;
  Proof pf;
  if(e[0] == e[1]) return reflexivityRule(e);
  Expr ite(e[0].iteExpr(e[1], d_em->trueExpr()));
  if(withProof()) {
    pf = newPf("ImpToIte", e);
  }
  return newRWTheorem(e, ite, a, pf); 
}


// ==> ITE(e, FALSE, TRUE) IFF NOT(e)
Theorem
CommonTheoremProducer::rewriteIteToNot(const Expr& e)
{
  if (CHECK_PROOFS)
    CHECK_SOUND(e.isITE() && e[1].isFalse() && e[2].isTrue(),
		"rewriteIteToNot: " + e.toString());
  Assumptions a;
  Proof pf;
  if (withProof()) pf = newPf("rewriteIteToNot", e);
  return newRWTheorem(e, e[0].negate(), a, pf);
}

// ==> ITE(a, TRUE, b) IFF OR(a, b)
Theorem
CommonTheoremProducer::rewriteIteToOr(const Expr& e)
{
  if (CHECK_PROOFS)
    CHECK_SOUND(e.isITE() && e[1].isTrue(),
		"rewriteIteToOr: " + e.toString());
  Assumptions a;
  Proof pf;
  if (withProof()) pf = newPf("rewriteIteToOr", e);
  return newRWTheorem(e, e[0] || e[2], a, pf);
}

// ==> ITE(a, b, FALSE) IFF AND(a, b)
Theorem
CommonTheoremProducer::rewriteIteToAnd(const Expr& e)
{
  if (CHECK_PROOFS)
    CHECK_SOUND(e.isITE() && e[2].isFalse(),
		"rewriteIteToAnd: " + e.toString());
  Assumptions a;
  Proof pf;
  if (withProof()) pf = newPf("rewriteIteToAnd", e);
  return newRWTheorem(e, e[0] && e[1], a, pf);
}

// ==> ITE(a, b, !b) IFF IFF(a, b)
Theorem
CommonTheoremProducer::rewriteIteToIff(const Expr& e)
{
  if (CHECK_PROOFS)
    CHECK_SOUND(e.isITE() && e[1] == e[2].negate(),
		"rewriteIteToIff: " + e.toString());
  Assumptions a;
  Proof pf;
  if (withProof()) pf = newPf("rewriteIteToIff", e);
  return newRWTheorem(e, e[0].iffExpr(e[1]), a, pf);
}

// ==> ITE(a, b, TRUE) IFF IMPLIES(a, b)
Theorem
CommonTheoremProducer::rewriteIteToImp(const Expr& e)
{
  if (CHECK_PROOFS)
    CHECK_SOUND(e.isITE() && e[2].isTrue(),
		"rewriteIteToImp: " + e.toString());
  Assumptions a;
  Proof pf;
  if (withProof()) pf = newPf("rewriteIteToImp", e);
  return newRWTheorem(e, e[0].impExpr(e[1]), a, pf);
}

///////////////////////////////////////////////////////////////////////
//// Conjunctive Normal Form (CNF) proof rules
///////////////////////////////////////////////////////////////////////
Theorem
CommonTheoremProducer::varIntroRule(const Expr& phi) {
  // This rule is sound for all expressions phi
  Assumptions a; Proof pf;
  const Expr boundVar = newExprBoundVar(phi);

  Expr body;  
  if(getType(phi).isBool())
    body = phi.iffExpr(boundVar);
  else
    body = phi.eqExpr(boundVar);

  std::vector<Expr> v; 
  v.push_back(boundVar);
  const Expr result = d_tm->getVCL()->existsExpr(v, body);
  
  if(withProof()) 
    pf = newPf("var_intro", phi);
  return newTheorem(result, a, pf);
}


Theorem
CommonTheoremProducer::andCNFRule(const Theorem& thm) {
  return opCNFRule(thm, AND, "and_cnf_rule");
}

Theorem
CommonTheoremProducer::orCNFRule(const Theorem& thm) {
  return opCNFRule(thm, OR, "or_cnf_rule");
}

Theorem
CommonTheoremProducer::impCNFRule(const Theorem& thm) {
  return opCNFRule(thm, IMPLIES, "implies_cnf_rule");
}

Theorem
CommonTheoremProducer::iffCNFRule(const Theorem& thm) {
  return opCNFRule(thm, IFF, "iff_cnf_rule");
}

Theorem
CommonTheoremProducer::iteCNFRule(const Theorem& thm) {
  return opCNFRule(thm, ITE, "ite_cnf_rule");
}


Theorem
CommonTheoremProducer::iteToClauses(const Theorem& ite) {
  const Expr& iteExpr = ite.getExpr();

  if(CHECK_PROOFS) {
    CHECK_SOUND(iteExpr.isITE() && getType(iteExpr).isBool(),
		"CommonTheoremProducer::iteToClauses("+iteExpr.toString()
		+")\n Argument must be a Boolean ITE");
  }
  const Expr& cond = iteExpr[0];
  const Expr& t1 = iteExpr[1];
  const Expr& t2 = iteExpr[2];

  Assumptions a;
  Proof pf;
  if(withAssumptions()) a = ite.getAssumptionsCopy();
  if(withProof())
    pf = newPf("ite_to_clauses", iteExpr, ite.getProof());
  return newTheorem((cond.negate() || t1) && (cond || t2), a, pf);
}


Theorem
CommonTheoremProducer::iffToClauses(const Theorem& iff) {
  if(CHECK_PROOFS) {
    CHECK_SOUND(iff.isRewrite() && getType(iff.getLHS()).isBool(),
		"CommonTheoremProducer::iffToClauses("+iff.getExpr().toString()
		+")\n Argument must be a Boolean IFF");
  }
  const Expr& t1 = iff.getLHS();
  const Expr& t2 = iff.getRHS();

  Assumptions a;
  Proof pf;
  if(withAssumptions()) a = iff.getAssumptionsCopy();
  if(withProof())
    pf = newPf("iff_to_clauses", iff.getExpr(), iff.getProof());
  return newTheorem((t1.negate() || t2) && (t1 || t2.negate()), a, pf);
}


//convert Predicates containing non-boolean ITEs into an ITE over
//the appropriate predicates
Theorem
CommonTheoremProducer::ifLiftRule(const Expr& e, int itePos) {
  if(CHECK_PROOFS) {
    CHECK_SOUND(getType(e).isBool(),
		"CommonTheoremProducer::ifLiftITERule("
		"input must be a Predicate: e = " + e.toString());
    CHECK_SOUND(e.arity() > itePos && e[itePos].isITE(),
		"CommonTheoremProducer::ifLiftITERule("
		"input does not have an ITE: e = " + e.toString());      
  }
  const Expr& ite = e[itePos];
  const Expr& cond = ite[0];
  const Expr& t1 = ite[1];
  const Expr& t2 = ite[2];

  if(CHECK_PROOFS) {
    CHECK_SOUND(getType(cond).isBool(),
		"CommonTheoremProducer::ifLiftITERule("
		"input does not have an ITE: e = " + e.toString());      
  }    

  vector<Expr> k1 = e.getKids();
  Op op(getOp(e));

  k1[itePos] = t1;
  Expr pred1 =  newExpr(op, k1);

  k1[itePos] = t2;
  Expr pred2 =  newExpr(op, k1);

  Expr resultITE = cond.iteExpr(pred1, pred2);

  Assumptions a; Proof pf;
  if (withProof())
    pf = newPf("ifLiftIteRule", e, newRatExpr(d_em, itePos));
  return newRWTheorem(e, resultITE, a, pf);
}

/////////////////////////////////////////////////////////////////////////
//// helper functions for CNF (Conjunctive Normal Form) conversion
/////////////////////////////////////////////////////////////////////////
Theorem
CommonTheoremProducer::opCNFRule(const Theorem& thm, 
				 int kind,
				 const string& ruleName) {
  TRACE("mycnf", "opCNFRule["+d_em->getKindName(kind)+"](",
	thm.getExpr(), ") {");
  ExprMap<Expr> localCache;
  if(CHECK_PROOFS) {
    Expr phiIffVar = thm.getExpr();
    CHECK_SOUND(phiIffVar.isIff(),
		"CommonTheoremProducer::opCNFRule("
		+d_em->getKindName(kind)+"): "
		"input must be an IFF: thm = " + phiIffVar.toString());
    CHECK_SOUND(phiIffVar[0].getKind() == kind,
		"CommonTheoremProducer::opCNFRule("
		+d_em->getKindName(kind)+"): "
		"input phi has wrong kind: thm = " + phiIffVar.toString());
    CHECK_SOUND(phiIffVar[0] != phiIffVar[1],
		"CommonTheoremProducer::opCNFRule("
		+d_em->getKindName(kind)+"): "
		"wrong input thm = " + phiIffVar.toString());   
    for(Expr::iterator it=phiIffVar[0].begin(), itend=phiIffVar[0].end();
	it!=itend;++it){
      CHECK_SOUND(phiIffVar[1] != *it,
		  "CommonTheoremProducer::opCNFRule("
		  +d_em->getKindName(kind)+"): "
		  "wrong input thm = " + phiIffVar.toString());
    }
  }
  const Expr& phi = thm.getExpr()[0];
  const Expr& phiVar = thm.getExpr()[1];

  std::vector<Expr> boundVars;
  std::vector<Expr> boundVarsAndLiterals;
  std::vector<Expr> equivs;
  for(Expr::iterator i=phi.begin(), iend=phi.end(); i != iend; i++) {
    // First, strip the negation and check if the formula is atomic
    Expr tmp(*i);
    while(tmp.isNot())
      tmp = tmp[0];

    if(d_core->isPropAtom(tmp))
      boundVarsAndLiterals.push_back(*i); 
    else
      boundVarsAndLiterals.push_back(findInLocalCache(*i, localCache,
						      boundVars));
  }
  
  for(ExprMap<Expr>::iterator it=localCache.begin(), itend=localCache.end(); 
      it != itend; it++) {
    DebugAssert((*it).second.isIff(),
		"CommonTheoremProducer::opCNFRule: " +
		(*it).second.toString());
    DebugAssert(!d_core->isPropLiteral((*it).second[0]) && 
		d_core->isLiteral((*it).second[1]),
		"CommonTheoremProducer::opCNFRule: " +
		(*it).second.toString());
    equivs.push_back((*it).second);
  }
  
  DebugAssert(boundVarsAndLiterals.size() == (unsigned)phi.arity(),
	      "CommonTheoremProducer::opCNFRule: "
	      "wrong size of boundvars: phi = " + phi.toString());
  
  DebugAssert(boundVars.size() == equivs.size(),
	      "CommonTheoremProducer::opCNFRule: "
	      "wrong size of boundvars: phi = " + phi.toString());
	      
  const Expr& cnfInput = newExpr(getOp(phi), boundVarsAndLiterals);
  Expr  result = convertToCNF(phiVar, cnfInput);
  if(boundVars.size() > 0)
    result = 
      d_tm->getVCL()->existsExpr(boundVars,
				 result.andExpr(andExpr(equivs)));
  
  Assumptions a;
  Proof pf;
  if(withAssumptions()) a = thm.getAssumptionsCopy();
  if(withProof())
    pf = newPf(ruleName, thm.getExpr(), thm.getProof());
  Theorem res(newTheorem(result, a, pf));
  TRACE("mycnf", "opCNFRule["+d_em->getKindName(kind)+"] => ", res.getExpr(),
	" }");
  return res;
}

//! produces the CNF for the formula  v <==> phi 
Expr CommonTheoremProducer::convertToCNF(const Expr& v, const Expr & phi) {
  //we assume that v \iff phi. v is the newVar corresponding to phi
  
  Expr::iterator i = phi.begin(), iend = phi.end();
  std::vector<Expr> clauses;
  std::vector<Expr> lastClause;
  switch(phi.getKind()) {
  case AND: {
    const Expr& negV = v.negate();
    lastClause.push_back(v);
    for(;i!=iend; ++i) {
      clauses.push_back(negV.orExpr(*i));
      lastClause.push_back(i->negate());
    }
    clauses.push_back(orExpr(lastClause));
  }
    break;
  case OR:{
    lastClause.push_back(v.negate());
    for(;i!=iend; ++i) {
      clauses.push_back(v.orExpr(i->negate()));
      lastClause.push_back(*i);
    }
    clauses.push_back(orExpr(lastClause));
  }
    break;
  case IFF: {
    const Expr& v1 = phi[0];
    const Expr& v2 = phi[1];
    Expr negV = v.negate();    
    Expr negv1 = v1.negate();
    Expr negv2 = v2.negate();
    clauses.push_back(newExpr(d_tm->getEM(),OR, negV, negv1, v2));
    clauses.push_back(newExpr(d_tm->getEM(),OR, negV, v1, negv2));
    clauses.push_back(newExpr(d_tm->getEM(),OR, v, v1, v2));
    clauses.push_back(newExpr(d_tm->getEM(),OR, v, negv1, negv2));
  } break;
  case IMPLIES:{
    const Expr& v1 = phi[0];
    const Expr& v2 = phi[1];
    Expr negV = v.negate();    
    Expr negv1 = v1.negate();
    Expr negv2 = v2.negate();
    clauses.push_back(newExpr(d_tm->getEM(),OR, negV, negv1, v2));
    clauses.push_back(v.orExpr(v1));
    clauses.push_back(v.orExpr(negv2));
  }
    break;
  case ITE: {
    const Expr& v1 = phi[0];
    const Expr& v2 = phi[1];
    const Expr& v3 = phi[2];
    const Expr& negV = v.negate();
    const Expr& negv1 = v1.negate();
    const Expr& negv2 = v2.negate();
    const Expr& negv3 = v3.negate();
    clauses.push_back(newExpr(d_tm->getEM(),OR, negV, negv1, v2));
    clauses.push_back(newExpr(d_tm->getEM(),OR, negV, v1, v3));
    clauses.push_back(newExpr(d_tm->getEM(),OR, v, negv1, negv2));
    clauses.push_back(newExpr(d_tm->getEM(),OR, v, v1, negv3));
  }
    break;
  default:
    DebugAssert(false, "CommonTheoremProducer::convertToCNF: "
		"bad operator in phi = "+phi.toString());
    break;
  }
  return andExpr(clauses);
}

///////////////////////////////////////////////////////////////////////
// helper functions for CNF converters
///////////////////////////////////////////////////////////////////////
/*! creates a new variable corresponding to the input boolean
 *  expression. this function is used in andCNFRules,orCNFRules,....
 */
Expr CommonTheoremProducer::newExprBoundVar(const Expr& e) {
  static int nextNum = 0;
//   DebugAssert(getType(e).isBool(),
// 	      "CommonTheoremProducer::newExprBoundVar: "
// 	      "input expression must have type boolean: " +
// 	      e.toString());
//   DebugAssert(e.isAnd() || e.isOr() || e.isImpl() || e.isIff() || e.isITE()
// 	      || e.isNot(),
// 	      "CommonTheoremProducer::newExprBoundVar: "
// 	      "wrong kind: e = " + e.toString());

  string name("_cnf_");
  std::string uid =  int2string(nextNum++);
  // IF_DEBUG(debugger.counter("CNF New Vars") = nextNum);
  Expr boundVar(newBoundVarExpr(d_em, name, uid));
  boundVar.setType(getType(e));
//   DebugAssert(d_core->isLiteral(boundVar) && !boundVar.isNot(),
// 	      "CommonTheoremProducer::newExprBoundVar: "
// 	      "incorrect entry in the cnf cache: " + boundVar.toString());
  return boundVar;
}

Expr 
CommonTheoremProducer::findInLocalCache(const Expr& i, 
					ExprMap<Expr>& localCache,
					vector<Expr>& boundVars) {
  TRACE("mycnf", "findInLocalCache(", i.toString(), ") { ");
  Expr boundVar;
  unsigned int negationDepth = 0;
  ExprMap<Expr>::iterator it;
 
  Expr phi = i;
  while(phi.isNot()) {
    phi = phi[0];
    negationDepth++;
  }
  
  it = localCache.find(phi);
  Expr v;
  if(it != localCache.end()) {
    v = ((*it).second)[1];
    IF_DEBUG(debugger.counter("CNF Local Cache Hits")++);
  }
  else {
    v = newExprBoundVar(i);
    boundVars.push_back(v);
    localCache[phi] = phi.iffExpr(v);
  }
  if(negationDepth % 2 != 0)
    v = !v;
  TRACE("mycnf", "findInLocalCache => ", v, " }");
  return v;
}
